/*
 * Copyright (c) 2003, 2004 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: lookupaddr.c,v 1.9 2004/08/15 16:45:33 ca Exp $")

#include "sm/assert.h"
#include "sm/error.h"
#include "sm/memops.h"
#include "sm/heap.h"
#include "map.h"
#include "sm/map.h"
#include "sm/maps.h"
#include "sm/mapc.h"

/*
**  Add flags (also for lookup_ip):
**  lookup also .domain
**  lookup only Tag:
**  others?
*/

/*
**  SM_MAP_LOOKUP_ADDR - Lookup RFC 2821 address
**
**	Parameters:
**		map -- map
**		user -- user part of address
**		detail -- detail part of address
**		domain -- domain part of address
**		tag -- tag including delimiter (NULL for no tag)
**		delim -- delimiter
**		flags -- flags to control lookup
**		rhs -- rhs of matching entry
**
**	Returns:
**		usual sm_error code
**	should it also return which case matched?
*/

#define UCH_AT		((uchar) '@')
#define UCH_STAR	((uchar) '*')
#define UCH_PLUS	((uchar) '+')

sm_ret_T
sm_map_lookup_addr(sm_map_P map, sm_str_P user, sm_str_P detail, sm_str_P domain, sm_str_P tag, uchar delim, uint32_t flags, sm_str_P rhs)
{
	sm_ret_T ret;
	size_t len, len_domain, j;
	uint32_t mapflags;
	sm_mapc_P mapc;
	sm_str_P str;

	if (map == NULL)
		return sm_error_perm(SM_EM_MAP, EINVAL);	/* XXX */
	SM_IS_MAP(map);
	mapc = map->sm_map_class;
	SM_IS_MAPC(mapc);
	if (mapc->sm_mapc_lookupf == NULL)
		return sm_error_perm(SM_EM_MAP, ENOENT);	/* XXX */

	len = 2;
	if (user != NULL)
		len += sm_str_getlen(user);
	if (detail != NULL)
		len += sm_str_getlen(detail) + 1;
	if (domain != NULL)
	{
		len_domain = sm_str_getlen(domain);
		len += len_domain + 1;
		--len_domain;
	}
	else
		len_domain = 0;
	if (tag != NULL)
		len += sm_str_getlen(tag);

	str = sm_str_new(NULL, len , len + 2);
	if (str == NULL)
		return sm_error_perm(SM_EM_MAP, ENOMEM);

	ret = sm_error_perm(SM_EM_MAP, SM_E_NOTFOUND);

/*
**  Macros to put '@' into the right place (after localpart or before domain)
**  depending on the flag SMMAP_LFL_NOAT.
*/

/* put an '@' as separator after localpart */
#define SM_AT_LOCAL	\
	(SMMAP_IS_LFL(flags, SMMAP_LFL_NOAT)	\
		? false : sm_is_err(sm_str_put(str, UCH_AT)))

/* put an '@' as separator before domain */
#define SM_AT_DOM	\
	(SMMAP_IS_LFL(flags, SMMAP_LFL_NOAT)	\
		? sm_is_err(sm_str_put(str, UCH_AT)) : false)

	j = 0;
	do
	{
		mapflags = SMMAP_FL_LWR_KEY
			| (tag != NULL ? SMMAP_FL_HAS_TAG : 0)
			| (user != NULL ? SMMAP_FL_HAS_LOCALPART : 0)
			| ((detail != NULL && sm_str_getlen(detail) > 0)
				? SMMAP_FL_HAS_DETAIL : 0)
			| (domain != NULL ? SMMAP_FL_HAS_DOMAIN : 0)
			;

		/* complete lookup (tag user delim detail @ domain) */
		if (SMMAP_IS_LFL(flags, SMMAP_LFL_FULL)
		    && SMMAP_LT_M_CAPS(map, mapflags))
		{
			sm_str_clr(str);
			if (
			      (tag != NULL && sm_is_err(sm_str_cat(str, tag)))
			   || (user != NULL && sm_is_err(sm_str_cat(str, user)))
			   || (detail != NULL && (sm_is_err(sm_str_put(str,
									delim))
					|| sm_is_err(sm_str_cat(str, detail))))
			   || SM_AT_LOCAL
			   || (domain != NULL &&
				(SM_AT_DOM ||
				 sm_is_err(sm_str_catpart(str, domain, j,
							len_domain))))
			   )
			{
				ret = sm_error_perm(SM_EM_MAP, EINVAL);	/* XXX */
				goto error;
			}
			ret = mapc->sm_mapc_lookupf(map, mapflags, str, rhs);
			if (ret == SM_SUCCESS)
				goto done;
		}

		/* lookup with ++ (tag user delim + @ domain) */
		if (SMMAP_IS_LFL(flags, SMMAP_LFL_DETPLUS)
		    && detail != NULL && sm_str_getlen(detail) > 0
		    && SMMAP_LT_M_CAPS(map, mapflags))
		{
			sm_str_clr(str);
			if (
			      (tag != NULL && sm_is_err(sm_str_cat(str, tag)))
			   || (user != NULL && sm_is_err(sm_str_cat(str, user)))
			   || (detail != NULL && (sm_is_err(sm_str_put(str,
									delim))
					|| sm_is_err(sm_str_put(str, UCH_PLUS))))
			   || SM_AT_LOCAL
			   || (domain != NULL &&
				(SM_AT_DOM ||
				 sm_is_err(sm_str_catpart(str, domain, j,
							len_domain))))
			   )
			{
				ret = sm_error_perm(SM_EM_MAP, EINVAL);	/* XXX */
				goto error;
			}
			ret = mapc->sm_mapc_lookupf(map, mapflags, str, rhs);
			if (ret == SM_SUCCESS)
				goto done;
		}

		/* lookup with +* (tag user delim * @ domain) */
		if (SMMAP_IS_LFL(flags, SMMAP_LFL_DETSTAR) && detail != NULL
		    && SMMAP_LT_M_CAPS(map, mapflags))
		{
			sm_str_clr(str);
			if (
			      (tag != NULL && sm_is_err(sm_str_cat(str, tag)))
			   || (user != NULL && sm_is_err(sm_str_cat(str, user)))
			   || (detail != NULL && (sm_is_err(sm_str_put(str,
									delim))
					|| sm_is_err(sm_str_put(str, UCH_STAR))))
			   || SM_AT_LOCAL
			   || (domain != NULL &&
				(SM_AT_DOM ||
				 sm_is_err(sm_str_catpart(str, domain, j,
							len_domain))))
			   )
			{
				ret = sm_error_perm(SM_EM_MAP, EINVAL);	/* XXX */
				goto error;
			}
			ret = mapc->sm_mapc_lookupf(map, mapflags, str, rhs);
			if (ret == SM_SUCCESS)
				goto done;
		}

		/* lookup without detail (tag user @ domain) */
		if (SMMAP_IS_LFL(flags, SMMAP_LFL_USER)
		    && SMMAP_LT_M_CAPS(map, mapflags))
		{
			mapflags = SMMAP_FL_LWR_KEY
				| (tag != NULL ? SMMAP_FL_HAS_TAG : 0)
				| (user != NULL ? SMMAP_FL_HAS_LOCALPART : 0)
				| ((detail != NULL && sm_str_getlen(detail) > 0)
					? SMMAP_FL_HAS_DETAIL : 0)
				| (domain != NULL ? SMMAP_FL_HAS_DOMAIN : 0)
				;
			sm_str_clr(str);
			if (
			      (tag != NULL && sm_is_err(sm_str_cat(str, tag)))
			   || (user != NULL && sm_is_err(sm_str_cat(str, user)))
			   || SM_AT_LOCAL
			   || (domain != NULL &&
				(SM_AT_DOM ||
				 sm_is_err(sm_str_catpart(str, domain, j,
							len_domain))))
			   )
			{
				ret = sm_error_perm(SM_EM_MAP, EINVAL);	/* XXX */
				goto error;
			}
			ret = mapc->sm_mapc_lookupf(map, mapflags, str, rhs);
			if (ret == SM_SUCCESS)
				goto done;
		}

/*
**  XXX This might not be the best order!
**	user@sub.domain - sub.domain - user@.domain - .domain
**  Wouldn't this be better:
**	user@sub.domain - user@.domain - sub.domain - .domain
*/
		mapflags = SMMAP_FL_LWR_KEY
			| (tag != NULL ? SMMAP_FL_HAS_TAG : 0)
			| (domain != NULL ? SMMAP_FL_HAS_DOMAIN : 0)
			;

		/* lookup just domain (@ domain) */
		if (SMMAP_IS_LFL(flags, SMMAP_LFL_DOMAIN)
		    && SMMAP_LT_M_CAPS(map, mapflags))
		{
			sm_str_clr(str);
			if (
			      (tag != NULL && sm_is_err(sm_str_cat(str, tag)))
			   || (domain != NULL &&
				(SM_AT_DOM ||
				 sm_is_err(sm_str_catpart(str, domain, j,
							len_domain))))
			   )
			{
				ret = sm_error_perm(SM_EM_MAP, EINVAL);	/* XXX */
				goto error;
			}
			ret = mapc->sm_mapc_lookupf(map, SMMAP_FL_LWR_KEY, str,
						rhs);
			if (ret == SM_SUCCESS)
				goto done;
		}

		if (domain != NULL && SMMAP_IS_LFL(flags, SMMAP_LFL_DOTSUBDOM))
		{
			++j;

			/* search for next '.' */
			while (j < len_domain
			       && sm_str_rd_elem(domain, j) != '.')
				++j;
		}

	} while (domain != NULL && SMMAP_IS_LFL(flags, SMMAP_LFL_DOTSUBDOM)
		 && j <= len_domain);

	/*
	**  This needs only to be done if domain is not NULL otherwise
	**  it's the same as above!
	*/

	if (domain != NULL && SMMAP_IS_LFL(flags, SMMAP_LFL_LOCAL))
	{
		mapflags = SMMAP_FL_LWR_KEY
			| (tag != NULL ? SMMAP_FL_HAS_TAG : 0)
			| (user != NULL ? SMMAP_FL_HAS_LOCALPART : 0)
			| ((detail != NULL && sm_str_getlen(detail) > 0)
				? SMMAP_FL_HAS_DETAIL : 0)
			;

		/* complete lookup (tag user delim detail) */
		if (SMMAP_IS_LFL(flags, SMMAP_LFL_FULL)
		    && SMMAP_LT_M_CAPS(map, mapflags))
		{
			sm_str_clr(str);
			if (
			      (tag != NULL && sm_is_err(sm_str_cat(str, tag)))
			   || (user != NULL && sm_is_err(sm_str_cat(str, user)))
			   || (detail != NULL && (sm_is_err(sm_str_put(str,
									delim))
					|| sm_is_err(sm_str_cat(str, detail))))
			   || SM_AT_LOCAL
			   )
			{
				ret = sm_error_perm(SM_EM_MAP, EINVAL);	/* XXX */
				goto error;
			}
			ret = mapc->sm_mapc_lookupf(map, mapflags, str, rhs);
			if (ret == SM_SUCCESS)
				goto done;
		}

		/* lookup with ++ (tag user delim +)*/
		if (SMMAP_IS_LFL(flags, SMMAP_LFL_DETPLUS)
		    && detail != NULL && sm_str_getlen(detail) > 0
		    && SMMAP_LT_M_CAPS(map, mapflags))
		{
			sm_str_clr(str);
			if (
			      (tag != NULL && sm_is_err(sm_str_cat(str, tag)))
			   || (user != NULL && sm_is_err(sm_str_cat(str, user)))
			   || (detail != NULL && (sm_is_err(sm_str_put(str,
									delim))
					|| sm_is_err(sm_str_put(str, UCH_PLUS))))
			   || SM_AT_LOCAL
			   )
			{
				ret = sm_error_perm(SM_EM_MAP, EINVAL);	/* XXX */
				goto error;
			}
			ret = mapc->sm_mapc_lookupf(map, mapflags, str, rhs);
			if (ret == SM_SUCCESS)
				goto done;
		}

		/* lookup with +* (tag user delim *) */
		if (SMMAP_IS_LFL(flags, SMMAP_LFL_DETSTAR) && detail != NULL
		    && SMMAP_LT_M_CAPS(map, mapflags))
		{
			sm_str_clr(str);
			if (
			      (tag != NULL && sm_is_err(sm_str_cat(str, tag)))
			   || (user != NULL && sm_is_err(sm_str_cat(str, user)))
			   || (detail != NULL && (sm_is_err(sm_str_put(str,
									delim))
					|| sm_is_err(sm_str_put(str, UCH_STAR))))
			   || SM_AT_LOCAL
			   )
			{
				ret = sm_error_perm(SM_EM_MAP, EINVAL);	/* XXX */
				goto error;
			}
			ret = mapc->sm_mapc_lookupf(map, mapflags, str, rhs);
			if (ret == SM_SUCCESS)
				goto done;
		}

		mapflags = SMMAP_FL_LWR_KEY
			| (tag != NULL ? SMMAP_FL_HAS_TAG : 0)
			| (user != NULL ? SMMAP_FL_HAS_LOCALPART : 0)
			;

		/* lookup without detail (tag user) */
		if (SMMAP_IS_LFL(flags, SMMAP_LFL_USER)
		    && SMMAP_LT_M_CAPS(map, mapflags))
		{
			sm_str_clr(str);
			if (
			      (tag != NULL && sm_is_err(sm_str_cat(str, tag)))
			   || (user != NULL && sm_is_err(sm_str_cat(str, user)))
			   || SM_AT_LOCAL
			   )
			{
				ret = sm_error_perm(SM_EM_MAP, EINVAL);	/* XXX */
				goto error;
			}
			ret = mapc->sm_mapc_lookupf(map, mapflags, str, rhs);
			if (ret == SM_SUCCESS)
				goto done;
		}
	}

  done:
	sm_str_free(str);
	return ret;

  error:
	sm_str_free(str);
	return ret;
}
