/*
 * Copyright (c) 2003-2005 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 *
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: t-mm.c,v 1.27 2005/05/01 02:09:08 ca Exp $")

#include "sm/error.h"
#include "sm/sysexits.h"
#include "sm/ctype.h"
#include "sm/heap.h"
#include "sm/memops.h"
#include "sm/maps.h"
#include "sm/mapc.h"
#include "sm/map.h"
#include "sm/mapclasses.h"
#include "sm/bdb.h"

#include "sm/io.h"

int Verbose = 0;

#define SEP	':'
#define MAPC_TYPE	"hash"
#define MAPC_NAME	"aliases"
#define MAPC_FILE	"aliases.db"
#define MAX_IN_LEN	(256 * 1024)

#ifndef SM_ALLOW_ROOT
# define SM_ALLOW_ROOT 0
#endif

static char buf[MAX_IN_LEN];

static int
makemap(sm_map_P map, int sep, uint32_t flags)
{
	sm_ret_T ret;
	sm_map_key_T key;
	sm_map_data_T data;
	char *delim, *rhs;
	int c, i, line;

	line = 0;
	while (fgets(buf, sizeof(buf), stdin) != NULL)
	{
		++line;
		delim = strchr(buf, sep);

		if (delim == NULL)
			continue;	/* complain about missing delim? */

		i = (delim - buf) + 1;
		*delim = '\0';

		/* skip over leading space; make this optional? */
		while (i < sizeof(buf) && (c = buf[i]) != '\0'
		       && isspace(c) && c != '\n')
			++i;
		if (i >= sizeof(buf) || c == '\0' || c == '\n')
		{
			fprintf(stderr,"rhs is empty, i=%d\n", i);
			continue;
		}
		rhs = buf + i;
		while (i < sizeof(buf) && (c = buf[i]) != '\0' && c != '\n')
			++i;
		if (i >= sizeof(buf) || c != '\n')
		{
			fprintf(stderr,"can't find end of rhs; line=%d\n",
				line);
			continue;
		}
		buf[i] = '\0';

		if ((flags & SMMAP_FL_LWR_KEY) != 0)
		{
			uint j;

			for (j = 0; j < strlen(buf); j++)
			{
				c = buf[j];
				if (ISUPPER(c))
					buf[j] = TOLOWER(c);
			}
		}

		sm_str_assign(key, NULL, (uchar *)buf, strlen(buf),
				strlen(buf));
		sm_str_assign(data, NULL, (uchar *)rhs, strlen(rhs),
				strlen(rhs));
		if (Verbose > 1)
		{
			fprintf(stderr,"lhs='%s' [%d], rhs='%s' [%d]\n"
				, (char *)sm_str_data(&key)
				, sm_str_getlen(&key)
				, (char *)sm_str_data(&data)
				, sm_str_getlen(&data)
				);
		}

		ret = sm_map_add(map, &key, &data, SMMAP_AFL_UNIQUE);
		if (sm_is_err(ret))
		{
			fprintf(stderr,"sm_map_add failed=%d; line=%d\n"
				, ret, line);
			sm_map_close(map, 0);
			return ret;
		}
	}

	ret = sm_map_close(map, 0);
	return ret;
}

static int
openmap(char *mapname, char *mapfile, int sep, uint32_t flags)
{
	sm_ret_T ret;
	sm_maps_P maps;
	sm_map_P map;
	sm_cstr_P mtype, mname;

	maps = NULL;
	mtype = mname = NULL;

	ret = sm_maps_init(&maps);
	if (maps == NULL)
		return ret;

	mtype = sm_cstr_scpyn0((const uchar *)MAPC_TYPE, strlen(MAPC_TYPE));
	if (mtype == NULL)
		goto error;

	mname = sm_cstr_scpyn0((const uchar *)mapname, strlen(mapname));
	if (mname == NULL)
		goto error;

	ret = sm_bdb_class_create(maps);

#if 0
	/* "delete" the file to make sure the result won't contain garbage? */
	ret = truncate(mapfile, 0);
#endif /* 0 */
	map = NULL;
	ret = sm_map_open(maps, mname, mtype, DB_CREATE, mapfile,
			SMAP_MODE_RDWR, &map, SMPO_END);
	if (!sm_is_success(ret))
		goto error;

	ret = makemap(map, sep, flags);

	ret = sm_maps_term(maps);
	SM_CSTR_FREE(mtype);
	SM_CSTR_FREE(mname);
	return ret;

  error:
	sm_maps_term(maps);
	return ret;
}

static sm_ret_T
usage(char *prg)
{
	fprintf(stderr, "usage: %s [options]\n"
		"-f: do not convert keys to lower case\n"
		"-F file: name of db file [%s]\n"
		"-n name: name of map [%s]\n"
		"-t c: separator [%c]\n"
		"\n%s creates a %s map from the data provided via stdin.\n"
		"Entries must be of the form\n"
		"lhs%crhs\n"
		"(where '%c' is the default separator)\n"
		"Entries cannot span multiple lines.\n"
		, prg
		, MAPC_FILE
		, MAPC_NAME
		, SEP
		, prg
		, MAPC_TYPE
		, SEP
		, SEP
		);
	return 1;
}


int
main(int argc, char *argv[])
{
	int c, sep;
	uint32_t flags;
	char *mapname, *mapfile, *prg;

	prg = argv[0];
	sep = (int) SEP;
	mapname = MAPC_NAME;
	mapfile = MAPC_FILE;
	flags = SMMAP_FL_LWR_KEY;
#if !SM_ALLOW_ROOT
	if (getuid() == 0 || geteuid() == 0)
	{
		fprintf(stderr,
			"%s: ERROR: do not run this as super-user!\n",
			prg);
		exit(EX_USAGE);
	}
#endif
	while ((c = getopt(argc, argv, "fF:hn:t:")) != -1)
	{
		switch (c)
		{
		  case 'f':
			flags &= ~SMMAP_FL_LWR_KEY;
			break;
		  case 'F':
			mapfile = optarg;
			break;
		  case 'n':
			mapname = optarg;
			break;
		  case 't':
			sep = (int) (*optarg);
			break;
		  case 'h':
		  default:
			usage(prg);
			return 1;
		}
	}
	c = openmap(mapname, mapfile, sep, flags);
	return c;
}
